import java.lang.reflect.Constructor;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;


public class EscapeXmlTest {

	public static void main(String[] args) {
		try {
			List<String> list = Files.readAllLines(Paths.get("placholder.html"), StandardCharsets.UTF_8);
			System.out.println("Warming up...");
			long t0 = System.currentTimeMillis();
			System.out.println("...nop");
			for (int i = 0; i < 10_000; i++) {
				nop(list);
			}
			System.out.println("...escapeXml_1");
			for (int i = 0; i < 10_000; i++) {
				test1(list);
			}
			System.out.println("...escapeXml_2");
			for (int i = 0; i < 10_000; i++) {
				test2(list);
			}
			System.out.println("...escapeXml_3");
			for (int i = 0; i < 10_000; i++) {
				test3(list);
			}
			System.out.println("...escapeXml_4");
			for (int i = 0; i < 10_000; i++) {
				test4(list);
			}
			System.out.println(System.currentTimeMillis() - t0);
			System.gc();
			System.out.println("Testing...");
			for (int x = 0; x < 20; x++) {
				System.out.println("...nop");
				t0 = System.currentTimeMillis();
				for (int i = 0; i < 1_000; i++) {
					nop(list);
				}
				System.out.println(System.currentTimeMillis() - t0);
				System.gc();
				System.out.println("...escapeXml_1");
				t0 = System.currentTimeMillis();
				for (int i = 0; i < 1_000; i++) {
					test1(list);
				}
				System.out.println(System.currentTimeMillis() - t0);
				System.gc();
				System.out.println("...escapeXml_2");
				t0 = System.currentTimeMillis();
				for (int i = 0; i < 1_000; i++) {
					test2(list);
				}
				System.out.println(System.currentTimeMillis() - t0);
				System.gc();
				System.out.println("...escapeXml_3");
				t0 = System.currentTimeMillis();
				for (int i = 0; i < 1_000; i++) {
					test3(list);
				}
				System.out.println(System.currentTimeMillis() - t0);
				System.gc();
				System.out.println("...escapeXml_4");
				t0 = System.currentTimeMillis();
				for (int i = 0; i < 1_000; i++) {
					test4(list);
				}
				System.out.println(System.currentTimeMillis() - t0);
				System.gc();
			}

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	private static String[] test4(List<String> list) {
		String[] result = new String[1000*list.size()];
		for (int i = 0; i < 1000; i++) {
			int x = 0;
			for (String str : list) {
				result[i*list.size() + x++] = escapeXml_4(str);
			}
		}
		return result;
	}

	private static String[] test3(List<String> list) {
		String[] result = new String[1000*list.size()];
		for (int i = 0; i < 1000; i++) {
			int x = 0;
			for (String str : list) {
				result[i*list.size() + x++] = escapeXml_3(str);
			}
		}
		return result;
	}

	private static String[] test2(List<String> list) {
		String[] result = new String[1000*list.size()];
		for (int i = 0; i < 1000; i++) {
			int x = 0;
			for (String str : list) {
				result[i*list.size() + x++] = escapeXml_2(str);
			}
		}
		return result;
	}

	private static String[] test1(List<String> list) {
		String[] result = new String[1000*list.size()];
		for (int i = 0; i < 1000; i++) {
			int x = 0;
			for (String str : list) {
				result[i*list.size() + x++] = escapeXml_1(str);
			}
		}
		return result;
	}

	private static String[] nop(List<String> list) {
		String[] result = new String[1000*list.size()];
		for (int i = 0; i < 1000; i++) {
			int x = 0;
			for (String str : list) {
				result[i*list.size() + x++] = str;
			}
		}
		return result;
	}
	
	/**
	 * XML escape input string
	 * 
	 * @param input
	 * @return
	 */
	public static String escapeXml_4(String input) {
		if (input == null || input.isEmpty()) {
			return input;
		}
		char[] out = null;
		int x = 0;
		final int ilen = input.length();
		for (int i = 0; i < ilen; i++) {
			final String repl;
			final char ch = input.charAt(i);
			switch (ch) {
				case '&': repl = "&amp;"; break;
				case '<': repl = "&lt;"; break;
				case '>': repl = "&gt;"; break;
				case '"': repl = "&#034;"; break;
				case '\'': repl = "&#039;"; break;
				default: repl = null; break;
			}
			if (repl != null) {
				if (out == null) {
					out = new char[xmllen(input)];
					if (i > 0) {
						//sb.append(charr, 0, i);
						//System.arraycopy(charr, 0, out, 0, i);
						input.getChars(0, i, out, 0);
					}
					x = i;
				}
				final int len = repl.length();
				repl.getChars(0, len, out, x);
				x += len;
			} else if (out != null) {
				out[x++] = ch;
			}
		}
		if (out != null) {
			return charArrayToString(out);
		}
		return input;
	}
	
	/**
	 * XML escape input string
	 * 
	 * @param input
	 * @return
	 */
	public static String escapeXml_3(String input) {
		if (input == null || input.isEmpty()) {
			return input;
		}
		char[] out = null;
		int x = 0;
		final char[] charr = input.toCharArray();
		for (int i = 0; i < charr.length; i++) {
			final String repl;
			final char ch = charr[i];
			switch (ch) {
				case '&': repl = "&amp;"; break;
				case '<': repl = "&lt;"; break;
				case '>': repl = "&gt;"; break;
				case '"': repl = "&#034;"; break;
				case '\'': repl = "&#039;"; break;
				default: repl = null; break;
			}
			if (repl != null) {
				if (out == null) {
					out = new char[xmllen(charr)];
					if (i > 0) {
						//sb.append(charr, 0, i);
						System.arraycopy(charr, 0, out, 0, i);
					}
					x = i;
				}
				final int len = repl.length();
				repl.getChars(0, len, out, x);
				x += len;
			} else if (out != null) {
				out[x++] = ch;
			}
		}
		if (out != null) {
			return charArrayToString(out);
		}
		return input;
	}

	/**
	 * XML escape input string
	 * 
	 * @param input
	 * @return
	 */
	public static String escapeXml_2(String input) {
		if (input == null || input.isEmpty()) {
			return input;
		}
		StringBuilder sb = null;
		for (int i = 0; i < input.length(); i++) {
			final String repl;
			final char ch = input.charAt(i);
			switch (ch) {
				case '&': repl = "&amp;"; break;
				case '<': repl = "&lt;"; break;
				case '>': repl = "&gt;"; break;
				case '"': repl = "&#034;"; break;
				case '\'': repl = "&#039;"; break;
				default: repl = null; break;
			}
			if (repl != null) {
				if (sb == null) {
					sb = new StringBuilder(xmllen(input));
					if (i > 0) {
						sb.append(input, 0, i);
					}
				}
				sb.append(repl);
			} else if (sb != null) {
				sb.append(ch);
			}
		}
		if (sb != null) {
			return sb.toString();
		}
		return input;
	}

	private static int xmllen(String str) {
		int len = str.length();
		for (int i = 0; i < str.length(); i++) {
			final char ch = str.charAt(i);
			switch (ch) {
				case '&': len += 4; break;
				case '<': len += 3; break;
				case '>': len += 3; break;
				case '"': len += 5; break;
				case '\'': len += 5; break;
			}
		}
		return len;
	}

	/**
	 * XML escape input string
	 * 
	 * @param input
	 * @return
	 */
	public static String escapeXml_1(String input) {
		if (input == null || input.isEmpty()) {
			return input;
		}
		StringBuilder sb = null;
		final char[] charr = input.toCharArray();
		for (int i = 0; i < charr.length; i++) {
			final String repl;
			final char ch = charr[i];
			switch (ch) {
				case '&': repl = "&amp;"; break;
				case '<': repl = "&lt;"; break;
				case '>': repl = "&gt;"; break;
				case '"': repl = "&#034;"; break;
				case '\'': repl = "&#039;"; break;
				default: repl = null; break;
			}
			if (repl != null) {
				if (sb == null) {
					sb = new StringBuilder(xmllen(charr));
					if (i > 0) {
						sb.append(charr, 0, i);
					}
				}
				sb.append(repl);
			} else if (sb != null) {
				sb.append(ch);
			}
		}
		if (sb != null) {
			return sb.toString();
		}
		return input;
	}

	private static int xmllen(char[] charr) {
		int len = charr.length;
		for (int i = 0; i < charr.length; i++) {
			final char ch = charr[i];
			switch (ch) {
				case '&': len += 4; break;
				case '<': len += 3; break;
				case '>': len += 3; break;
				case '"': len += 5; break;
				case '\'': len += 5; break;
			}
		}
		return len;
	}

	private static final Constructor<String> SECRET_STRING_CTOR;

	static {
		Constructor<String> secret = null;
		try {
			Constructor<String> c = String.class.getDeclaredConstructor(char[].class, boolean.class);
			c.setAccessible(true);
			String s = c.newInstance(new char[]{'c'}, true);
			if ("c".equals(s)) {
				secret = c;
			}
		} catch (Exception e) {
		}
		SECRET_STRING_CTOR = secret;
	}
	
	private static final String charArrayToString(char[] chars) {
		if (SECRET_STRING_CTOR != null) {
			try {
				return SECRET_STRING_CTOR.newInstance(chars, true);
			} catch (Exception ignored) {
			}
		}
		return new String(chars);
	}
}
